import numpy as np
import pytest
from scipy import stats

from epistoch.utils.stats import constant, get_lognorm, loss_function
from pyphase.phase import ph_erlang


def test_constant():
    v = constant()
    assert 0 == v.mean()
    v = constant(loc=8)
    assert 8 == v.mean()
    const_loss = loss_function(v)
    assert 1 == const_loss(7)


def test_loss_function():
    EPS = 1e-5
    expo = stats.expon(scale=4)
    gm1 = stats.gamma(a=1, scale=4)  # identical to previous one
    gm2 = stats.gamma(a=2, scale=3)
    nrm = stats.norm(loc=4, scale=1)
    lnrm = get_lognorm(4, 1)
    const = constant(loc=4)
    for force in [True, False]:
        loss_expo = loss_function(expo, force=force)
        loss_gm1 = loss_function(gm1, force=force)
        loss_gm2 = loss_function(gm2, force=force)
        loss_norm = loss_function(nrm, force=force)
        loss_lnrm = loss_function(lnrm, force=force)
        loss_const = loss_function(const, force=force)
        assert 4 == pytest.approx(loss_expo(0))
        assert 4 == pytest.approx(loss_gm1(0))
        assert 4 == pytest.approx(loss_lnrm(0))
        assert 4 == pytest.approx(loss_const(0))
        x = np.array([-2, -1, 0, 1, 3, 1000])
        # For positive vars loss1(0) = mean
        loss_expo_at_x = np.array(
            [6.00000000e000, 5.00000000e000, 4.00000000e000, 3.11520313e000, 1.88946621e000, 1.06882573e-108]
        )
        np.testing.assert_allclose(loss_expo(x), loss_expo_at_x, atol=EPS, err_msg=f"Expo loss, force={force}")
        np.testing.assert_allclose(loss_gm1(x), loss_expo_at_x, atol=EPS, err_msg=f"Gamma1 loss, force={force}")
        loss_gamma2_at_x = np.array(
            [8.00000000e000, 7.00000000e000, 6.00000000e000, 5.01571917e000, 3.31091497e000, 1.72896654e-142]
        )
        np.testing.assert_allclose(loss_gm2(x), loss_gamma2_at_x, atol=EPS)
        loss_norm_at_x = np.array([6.0, 5, 4, 3.00038215, 1.08331547, 0.0])
        assert loss_norm_at_x == pytest.approx(loss_norm(x), rel=EPS)
        loss_lnrm_at_x = np.array([6, 5, 4, 3.00000000e000, 1.05078013e000, 7.72290351e-112])
        assert loss_lnrm_at_x == pytest.approx(loss_lnrm(x), rel=EPS)
        loss_const_at_x = np.array([6.0, 5.0, 4.0, 3.0, 1.0, 0.0])
        np.testing.assert_allclose(loss_const(x), loss_const_at_x, err_msg=f"Constant loss, force={force}")


def test_ph_loss():
    phd = ph_erlang(3, 10)
    gma = stats.gamma(a=3, scale=1 / 10)
    x = np.linspace(0, 2)
    loss_gma = loss_function(gma, force=False)
    loss_gma_at_x = loss_gma(x)
    loss_ph_at_x = phd.loss1(x)
    np.testing.assert_allclose(loss_gma_at_x, loss_ph_at_x)


if __name__ == "__main__":
    test_constant()
    test_loss_function()
    test_ph_loss()



# -*- coding: utf-8 -*-
"""
Created on Sat Apr 11 12:20:32 2020

@author: Germán Riaño (griano@germanriano.com)
"""


import logging
import math

import numpy as np
import pandas as pd
from scipy import integrate, interpolate, optimize, stats
from tqdm import tqdm

from epistoch.utils.stats import loss_function

EPS = 1e-5


# The SIR model differential equations.
def _sir_deriv(t, y, beta, gam):
    S, I = y
    dSdt = -beta * S * I
    dIdt = beta * S * I - gam * I
    return dSdt, dIdt


def sir_classical(
    name, population, reproductive_factor, infectious_period_mean, num_days, I0=1.0, S0=None, use_odeint=True,
):
    """
    Solves a classical SIR model.

    Parameters
    ----------
    name: string
        Model name
    population: float
        total population size
    reproductive_factor: float
        basic reproductive factor(R0)
    infectious_period_mean: float
        expected value of Infectious Period Time
    I0: float
        Initial infectious population
    S0: float
        Initial susceptible population (optinal, defaults to all but I0)
    num_days: int
        number of days to run


    Returns
    -------
    dict
        Dictionary with fields:
            - name: model name
            - population: Total population
            - total_infected
            - data: data Frame with columns
                - S : Susceptible,
                - I : Infectious (Dying or recovering),
                - R : Removed (recovered + deaths),
    """
    # Note: Code based on
    # https://scipython.com/book/chapter-8-scipy/additional-examples/the-sir-epidemic-model/
    # Total population, N.
    N = population
    # Initial number of infected and recovered individuals, I0 and R0.
    I0 = I0 / N
    S0 = S0 / N if S0 is not None else 1 - I0
    # Note that R0 = 1 - I0 - S0

    # Contact rate, beta, and mean recovery rate, gam, (in 1/days).
    gam = 1.0 / infectious_period_mean
    beta = reproductive_factor * gam

    # A grid of time points (in days)
    times = np.linspace(0, num_days, num_days + 1)

    # Initial conditions vector
    y0 = [S0, I0]
    # Integrate the SIR equations over the time grid, t.
    ret = integrate.odeint(_sir_deriv, y0, times, args=(beta, gam), tfirst=True)
    S, I = ret.T
    # De-normalize
    S = S * N
    I = I * N
    R = N - S - I
    result = dict()
    result["data"] = pd.DataFrame(data={"Day": times, "S": S, "I": I, "R": R}).set_index("Day")
    result["total_infected"] = get_total_infected(reproductive_factor)
    result["population"] = N
    result["name"] = name
    return result


def compute_integral(n, delta, S, I, times, survival, pdfs, loss1, dist, method="loss"):
    r"""
    Compute the integral needed for the integro-differential model.

    In other words, computes

    .. math:: \int_0^t g(t-x) I(x)S(x) dx \quad  \text{ for } t = n*delta

    Parameters
    ----------
    n : integer
        upper limit for integral.
    delta : float
        interval size
    survival : array of float
        array G_k = P( T > delta k)
    delta_loss : array of float
        array L_k = E(T-delta*k)

    Returns
    -------
    Value of the integral

    """
    if n == 0:
        return 0.0
    if method == "loss":
        IS = np.zeros_like(survival)
        # next line equivalent to: IS[: n + 1] = np.array([I[n - k] * S[n - k] for k in range(n + 1)])
        IS[: n + 1] = np.flipud(I[: n + 1] * S[: n + 1])
        slopes = np.diff(IS, append=0.0) / delta  # m1, m2, ...
        delta_slopes = np.diff(slopes, prepend=0.0)
        return IS[0] + sum(delta_slopes * loss1)

    if method == "simpson":
        integral_points = [pdfs[n - k] * S[k] * I[k] for k in range(0, n + 1)]
        return integrate.simps(integral_points, dx=delta)
    if method == "interpolate":
        t = n * delta
        interpolator = interpolate.interp1d(times[: n + 1], S[: n + 1] * I[: n + 1])

        def integrand(tau):
            return dist.pdf(t - tau) * interpolator(tau)

        return integrate.quad(integrand, 0, t)[0]


def sir_g(
    name,
    population,
    reproductive_factor,
    infectious_time_distribution,
    num_days,
    I0=1.0,
    S0=None,
    num_periods=None,
    method="loss",
    logger=None,
):
    """
    Solves a SIR-G model.

    Parameters
    ----------
    name: string
        Model name
    population: float
        total population size
    reproductive_factor: float
        basic reproductive factor(R0)
    infectious_time_distribution: scipy.stats.rv_continuous
        expected value of Infectious Period Time
    num_days: int
        number of days to run
    I0: float
        Initial infectious population
    S0: float
        Initial susceptible population (optional, defaults to all but I0)
    num_periods:int
        Number of periods to use for computations. Higher number will lead ot more precise computation.
    method: string
        Method used for the internal integral
    logger
        Logger object

    Returns
    -------
    dict
        Dictionary with fields:
            - name: model name
            - population: Total population
            - total_infected
            - data: data Frame with columns
                - S : Susceptible,
                - I : Infectious (Dying or recovering),
                - R : Removed (recovered + deaths),
    """
    if logger is None:
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.INFO)
    # Total population, N.
    N = population
    dist = infectious_time_distribution

    I0 = I0 / N
    S0 = S0 / N if S0 is not None else 1 - I0
    # Notice that R0 = 1 - I0 - S0

    # Contact rate, beta, and mean recovery rate, gam, (in 1/days).
    gam = 1 / dist.mean()
    beta = reproductive_factor * gam
    # A grid of time points (in days)
    num_periods = 10 * num_days if num_periods is None else num_periods
    delta = num_days / num_periods
    times = np.linspace(start=0.0, stop=num_days, num=num_periods + 1)
    logging.info("Computing survival function")
    survival = dist.sf(times)
    logging.info("Computing PDF function")
    pdfs = dist.pdf(times)
    if pdfs[0] == math.inf:
        pdfs[0] = dist.pdf(dist.ppf(0.01))
    survival = dist.sf(times)
    if survival[0] < 1.0 - EPS:
        logging.warning(f"Survival function does not start in 1:  {survival}")
    logging.info("Computing Loss1 function")
    loss_fun = loss_function(dist)
    loss1 = loss_fun(times)

    S = np.zeros(num_periods + 1)
    # S[:] = np.nan
    I = np.zeros(num_periods + 1)
    #    I[:] = np.nan
    S[0] = S0
    I[0] = I0
    logging.info(f"Computing SIR-G model with {num_periods} periods.")
    for n in tqdm(range(0, num_periods)):
        S[n + 1] = S[n] - delta * beta * S[n] * I[n]
        integral = compute_integral(n, delta, S, I, times, survival, pdfs, loss1, dist, method=method)
        # Expected case for exponential variables
        integral_expon = (gam / beta) * (I[n] - I0 * survival[n])
        logger.debug(f"n={n}, t={n*delta}, integral = {integral}, integral_expon = {integral_expon}")
        I[n + 1] = I[n] + delta * (beta * (S[n] * I[n] - integral) - gam * I0 * survival[n])

    # Report one point per day, and de-normalize
    days = np.linspace(0, num_days, num_days + 1)
    S = N * interpolate.interp1d(times, S)(days)
    I = N * interpolate.interp1d(times, I)(days)
    R = N - S - I
    result = dict()
    result["name"] = name
    result["population"] = population
    result["data"] = pd.DataFrame(data={"Day": days, "S": S, "I": I, "R": R}).set_index("Day")
    result["total_infected"] = get_total_infected(reproductive_factor)
    return result


def get_total_infected(reproductive_factor, normalzed_s0=1):
    r"""
    Estimate the total number of infected for a given reproductive factor.

    Find the value z that solves

    ..math 1-z = S_0 * e^{{\mathcal R} z}

    Parameters
    ----------
    reproductive_factor : float
        Basic reproductive factor.
    normalzed_s0 : float, optional
        Initial fraction of population that is infected. The default is 1.

    Returns
    -------
    float
        The stimated fraction of toal infected.

    """
    if reproductive_factor < 1:
        return 0.0
    else:
        fun = lambda x: 1 - x - normalzed_s0 * np.exp(-reproductive_factor * x)
        result = optimize.root_scalar(fun, bracket=[0.0001, 1])
        return result.root


def _compute_array_error(name, x1, x2, N, do_print=True):
    # Convenience method to compute the error difference in tow array
    error = np.abs(x1 - x2) / N
    if do_print:
        print(f"{name}: max error = {np.max(error):.2}, avg error = {np.mean(error):.2}")
    return np.max(error)


def _compute_error(model1, model2, N, do_print=True):
    # Convenience methods to compute errors between two models
    if model1["population"] != model2["population"]:
        raise ValueError("We can compare only models with the same population")
    N = model1["population"]
    error_i = _compute_array_error("I", model1["data"].I, model2["data"].I, N, do_print)
    error_s = _compute_array_error("S", model1["data"].S, model2["data"].S, N, do_print)
    return 0.5 * (error_i + error_s)


def print_error(model1, model2):
    """
    Print error difference between two models

    Parameters
    ----------
    model1
        Result of a SIR, SIR_G model
    model2
        Result of a SIR, SIR_G model

    Returns
    -------

    """
    return _compute_error(model1, model2, True)


def report_summary(result):
    """
    Report a summary for a model.

    Parameters
    ----------
    result:
        Result from SIR like model
    """
    model = result["data"]
    name = result["name"]
    N = result["population"]
    n = len(model) - 1
    print(f"Model {name} Summary")
    print(f"  Total Infected People: {int(model.R[n]):,d} ({model.R[n]/N:.2%})")
    print(f"  Infection Peak: {int(np.max(model.I)):,d} ({np.max(model.I)/N:.2%})")
    print(f"  Peak Day: {int(np.argmax(model.I)):,d}")
    if "total_infected" in result:
        print(f"  Theoretical Total Infected: {result['total_infected']:.2%}")



# -*- coding: utf-8 -*-
"""
Created on Tue Apr 21 17:14:05 2020

@author: Germán Riaño
"""
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick

plt.rc("text", usetex=True)
plt.rc("font", **{"family": "serif", "sans-serif": ["Palatino"]})
plt.set_loglevel("info")


def prepare_plot(title):
    fig = plt.figure(facecolor="w")
    # use LaTeX fonts in the plot

    # ax = fig.add_subplot(111, facecolor='#dddddd', axisbelow=True) # Gray foreground
    ax = fig.add_subplot(111, axisbelow=True)
    ax.set_xlabel("Time /days")
    ax.set_ylabel("Population Percent")
    # ax.set_ylim(0,1.1)
    ax.grid(b=True, which="major", c="k", lw=0.5, ls=":")
    ax.set_title(title)
    ax.yaxis.set_major_formatter(mtick.PercentFormatter(xmax=1, decimals=0))
    return fig, ax


def plot_sir(
    model,
    fig=None,
    title=None,
    formats={"S": "b-", "I": "r-", "R": "g-"},
    legend_fmt={"loc": "best", "shadow": True},
    **kwd,
):
    data = model["data"]
    name = model["name"]
    N = model["population"]
    # Plot the data on three separate curves for S(t), I(t) and R(t)
    if fig is None:
        title = name if title is None else title
        fig, ax = prepare_plot(title)
    else:
        allaxes = fig.get_axes()
        ax = allaxes[0]
    t = data.index
    for s, fmt in formats.items():
        series = data[s] / N
        linewidth = kwd.get("linewidth", 2)
        linestyle = kwd.get("linestyle", "-")
        ax.plot(
            t, series, fmt, alpha=0.8, lw=linewidth, linestyle=linestyle, label=f"{s} - {name}",
        )
    ax.legend(**legend_fmt)
    return fig


def plot_IR(model, fig=None, linestyle="-", title=None, legend_fmt={"loc": "best", "shadow": True}):
    # Plot the data on three separate curves for S(t), I(t) and R(t)
    data = model["data"]
    name = model["name"]
    N = model["population"]
    if fig is None:
        title = name if title is None else title
        fig, ax = prepare_plot(title)
    else:
        allaxes = fig.get_axes()
        ax = allaxes[0]
    ax.plot(
        data.S / N, data.R / N, "g", alpha=0.8, lw=2, linestyle=linestyle, label="Removed - " + name,
    )
    ax.plot(
        data.S / N, data.I / N, "b", alpha=0.8, lw=2, linestyle=linestyle, label="Infected - " + name,
    )
    ax.legend(**legend_fmt)
    return fig



# -*- coding: utf-8 -*-
"""
Created on Wed Apr 15 15:14:55 2020

@author: Germán Riaño
"""

import math
import warnings

import numpy as np
from scipy import integrate, stats


def get_lognorm(mu, sigma):
    """
    Builds a lognormal distribution from mean and standard deviation for this 
    variable. (Not the mean and sd of the corresponding normal)

    Parameters
    ----------
    mu : double
        The expected value
    sigma : double
        standard deviation

    Returns
    -------
    A frozen ditribution object

    """
    sigma2 = sigma * sigma
    mu2 = mu * mu
    ln_mu = np.log(mu2 / np.sqrt(mu2 + sigma2))
    ln_sigma = np.sqrt(np.log(1 + sigma2 / mu2))
    return stats.lognorm(s=ln_sigma, scale=math.exp(ln_mu))


def get_gamma(mu, sigma):
    """
    Builds a gamma distribution from mean and standard deviation for this 
    variable.

    Parameters
    ----------
    mu : double
        The expected value
    sigma : double
        standard deviation

    Returns
    -------
    A frozen ditribution object

    """
    alpha = (mu / sigma) ** 2
    beta = mu / alpha
    return stats.gamma(a=alpha, scale=beta)


class ConstantDist(stats.rv_continuous):
    def _cdf(self, x):
        return np.where(x >= 0, 1.0, 0.0)

    def _ppf(self, p):
        return 0.0

    def _loss1(self, x):
        return np.where(x > 0, 0.0, -x)  # [E(0-x)^+]


constant = ConstantDist(a=0.0, name="constant")


def loss_function(dist, force=False):

    """
    Creates a loss function of order 1 for a distribution from scipy

    Parameters
    ----------
    dist : scipy.stats._distn_infrastructure.rv_froze
        a distribution object form scipy.stats
    force : boolean
        whether force an integral computation instead of known formula        

    Returns
    -------
    Callable that represent this loss function

    """
    lo, hi = dist.support()
    loc, scale = None, None
    if not force:
        if "loss1" in dir(dist):
            return dist.loss1
        name = None
        return_fun = None
        if "name" in dir(dist):
            name = dist.name
        elif "dist" in dir(dist):
            if "name" in dir(dist.dist):
                name = dist.dist.name
        if name == "expon":
            return_fun = lambda x: np.exp(-x)
        if name == "gamma":
            a = dist.kwds["a"]
            return_fun = lambda x: a * stats.gamma.sf(x, a=a + 1) - x * stats.gamma.sf(x, a)
        # Standard normal loss function used below
        if name == "norm":
            # loc and scale not set for the normal
            loc = dist.args[0] if len(dist.args) > 1 else None
            scale = dist.args[1] if len(dist.args) > 2 else None
            return_fun = lambda z: stats.norm.pdf(z) - z * stats.norm.sf(z)
        if name == "lognorm":

            def loss1(x):
                s = dist.kwds["s"]
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore")
                    result = np.exp(0.5 * s * s) * stats.norm.sf(np.log(x) / s - s) - x * stats.norm.sf(np.log(x) / s)
                return result

            return_fun = loss1
        if name == "constant":
            loc = dist.args[0] if len(dist.args) > 0 else None
            return_fun = dist.dist._loss1
        if return_fun is not None:
            loc = dist.kwds.get("loc", 0.0) if loc is None else loc
            scale = dist.kwds.get("scale", 1.0) if scale is None else scale
            loss1 = lambda x: np.where(x > lo, scale * return_fun((x - loc) / scale), dist.mean() - x)
            return loss1

    loss1 = np.vectorize(lambda x: integrate.quad(dist.sf, x, hi)[0])
    return loss1


def avg_recovery_rate(dist):
    loss = loss_function(dist)

    def integrand(t):
        sf = dist.sf(t)
        lss = loss(t)
        result = np.zeros_like(t)
        valid = lss > 0
        result[valid] = sf[valid] * sf[valid] / lss[valid]
        return result

    gam = 1.0 / dist.mean()
    a, b = dist.support()
    result = gam * integrate.quad(integrand, a, b)[0]
    return result



# -*- coding: utf-8 -*-
"""
Created on Sun Apr 26 11:43:11 2020

@author: Germán Riaño, PhD
"""
import logging

import numpy as np
import pandas as pd
from numpy import matlib as ml
from scipy import integrate, interpolate, linalg

from epistoch.sir_g import get_total_infected
from pyphase.phase import ph_expon

EPS = 1e-5


def to_array(*args):
    flat_args = [np.array(arg).flatten() for arg in args]
    return np.concatenate(flat_args)


# The SIR-PH model differential equations.
def _deriv(t, y, beta, n, alpha, A, a):
    S, x, r = np.split(y, [1, n + 1])
    x = ml.mat(x)
    x.reshape(1, n)
    x_beta = x @ beta
    dSdt = -S * x_beta
    dxdt = S * (x @ beta @ alpha) + x @ A
    drdt = np.multiply(x.T, a)  # component-wise multiplication
    return to_array(dSdt, dxdt, drdt)


def sir_phg(
    name,
    population,
    beta,
    infectious_time_distribution,
    num_days=160,
    I0=1.0,
    S0=None,
    logger=None,
    report_phases=False,
):
    """
    Compute a SIR-PH model

    Parameters
    ----------
    name: string
        Model name
    population : float, optional
        Total population. The default is 1000.
    beta : float, optional
        Contagion rate. The default is 0.2.
    infectious_time_distribution : phase, optional
        Must be a PH dstribution. The default is ph_expon(lambd=1 / 10).
    num_days : int
        Number of days to run.
    I0 : float, optional
        initial population. The default is 1.0.
    S0 : float or int, optional
        Initial susceptible. The default is all but I0.
    logger : Logger, optional
        Logger object. If none is given, default logging used.
    report_phases : boolean, optional
        Whether to include phase levels in the report. The default is False.

    Returns
    -------
    A dictionary with these fields:

        data: DataFrame with columns S, I, R. Optionally: I-Phase0,...,I-Phase(n-1), and R-Phase0,...R-Phase(n-1)

        total_infected: estimation of total infected individuals

        name: model name

        population: total population
    """
    if logger is None:
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.INFO)
    # Total population, N.
    N = population
    dist = infectious_time_distribution

    I0 = I0 / N
    S0 = S0 / N if S0 is not None else 1 - I0
    R0 = 1 - I0 - S0

    # Contact rate, beta, and mean recovery rate, gam, (in 1/days).
    gam = 1 / dist.mean()
    # A grid of time points (in days)
    times = np.linspace(start=0.0, stop=num_days, num=num_days + 1)

    alpha, A, a, n = dist.params()
    beta = beta * np.ones(n)

    # Spread initial I in phases
    x0 = -I0 * gam * alpha @ linalg.inv(A)
    # Spread initial R0 evenly  in phases
    r0 = R0 * np.ones(n) / n

    logging.info(f"Computing SIR-PH model with {num_days} days and {n} phases.")
    # Initial conditions vector
    y0 = to_array(S0, x0, r0)
    # Integrate the SIR-PH equations over the time grid, t.
    ret = integrate.odeint(_deriv, y0, times, args=(beta, n, alpha, A, a), tfirst=True)
    S, x, r = np.split(ret, [1, n + 1], axis=1)

    S = S.flatten()
    I = np.sum(x, axis=1)
    R = np.sum(r, axis=1)
    pi = dist.equilibrium_pi()
    effective_r0 = (pi @ beta) / gam

    # Report one point per day, and de-normalize
    days = np.linspace(0, num_days, num_days + 1)
    S = N * interpolate.interp1d(times, S)(days)
    I = N * interpolate.interp1d(times, I)(days)
    R = N * interpolate.interp1d(times, R)(days)
    result = dict()
    result["data"] = pd.DataFrame(data={"Day": days, "S": S, "I": I, "R": R}).set_index("Day")
    if report_phases:
        xdf = pd.DataFrame(N * x)
        xdf.columns = ["I-Phase" + str(i) for i in range(n)]
        result["data"] = pd.concat([result["data"], xdf], axis=1)
        result["I-columns"] = xdf.columns.values.tolist()
        rdf = pd.DataFrame(N * r)
        rdf.columns = ["R-Phase" + str(i) for i in range(n)]
        result["data"] = pd.concat([result["data"], rdf], axis=1)
        result["R-columns"] = rdf.columns.values.tolist()
    result["total_infected"] = get_total_infected(effective_r0, S0)
    result["name"] = name
    result["population"] = population

    return result



import numpy as np
from numpy import matlib as ml
from scipy.stats import expon, gamma

from pyphase.phase import ph_erlang, ph_expon, ph_mix, ph_sum, phase
from tests.pyphase_test.testing import assert_dist_identical, mix_rv


def test_ph_expo():
    lambd = 10
    exp = expon(scale=1.0 / lambd)
    dists = dict()
    dists["np"] = phase(alpha=np.array([1.0]), A=np.array([-lambd]))
    dists["float"] = phase(alpha=1.0, A=-lambd)
    dists["mat"] = phase(alpha=ml.mat([1.0]), A=ml.mat([[-lambd]]))
    dists["method"] = ph_expon(lambd)
    for name, dist in dists.items():
        assert_dist_identical(exp, dist, "Expo-" + name)


def test_ph_gamma():
    lam = 20.0
    n = 2
    gam = gamma(a=n, scale=1 / lam)
    a = [1.0, 0.0]
    A = [[-lam, lam], [0, -(lam)]]
    dists = dict()
    dists["np"] = phase(alpha=np.array(a), A=np.array(A))
    dists["mat"] = phase(alpha=ml.mat(a), A=ml.mat(A))
    dists["method"] = ph_erlang(2, lam)
    for name, dist in dists.items():
        assert_dist_identical(gam, dist, "Gamma-" + name)


def test_ph_sum():
    lam = 3.0
    v1 = phase(alpha=np.array([1.0]), A=np.array([-lam]))
    v2 = v1
    v = ph_sum(v1, v2)
    print(f"sum = {v}")
    expected = gamma(a=2, scale=1 / lam)
    assert_dist_identical(expected, v, "Sum Expos")


def test_ph_mix():
    lam1 = 4.0
    lam2 = 2.0
    p = 0.4
    v1 = ph_expon(lam1)
    v2 = ph_expon(lam2)
    v = ph_mix(v1, v2, p)
    print(f"mix = {v}")
    expected = mix_rv(ps=np.array([p, 1 - p]), vs=[v1, v2])
    x = np.array([1, 2, 3])
    print(f"v-cdf = {v.cdf(x)}")
    print(f"ex-cdf = {expected.cdf(x)}")
    print(f"v-moms: {[{n:v.moment(n) for n in range(1,4)}]}")
    print(f"exp-moms: {[{n:expected.moment(n) for n in range(1,4)}]}")
    assert_dist_identical(expected, v, "Mix-HyperExpo")


def test_loss1():
    lambd = 0.5
    v1 = ph_expon(lambd)
    x = np.linspace(1, 5, 1)
    exp_loss1 = lambda x: np.exp(-lambd * x) / lambd  # Noqa 731
    np.testing.assert_allclose(v1.loss1(x), exp_loss1(x))


def test_equilibrium_phases():
    v = ph_erlang(n=3, mean=3)
    print(v)
    pi = v.equilibrium_pi()
    print(pi)


#######################################
#  T E S T I N G
#######################################


if __name__ == "__main__":
    test_equilibrium_phases()
    test_ph_expo()
    test_ph_gamma()
    test_loss1()
    test_ph_sum()
    test_ph_mix()



# -*- coding: utf-8 -*-
"""
Created on Sun Apr 26 21:50:11 2020

@author: Germán Riaño, PhD
"""

import numpy as np

from epistoch.sir_phg import sir_phg
from pyphase.phase import ph_erlang, ph_mix, ph_sum


def seird_ph(
    name, population, beta, exposed_time, time_to_die, time_to_recover, fatality_rate, num_days, I0=1, logger=None,
):
    """
    Compute a SEIRD model with Phase-Type distribution for the different stages.

    Parameters
    ----------
    name : string
        Model name.
    population : int or float
        Population size.
    beta : float
        Contagion rate.
    exposed_time : PH distribution
        Distribution of time exposed, not yet infectious.
    time_to_die : PH distribution
        Time to die after becoming infectious.
    time_to_recover : PH distribution
        Time to recover after becoming infectious.
    fatality_rate : float
        Percentage of individuals that die.
    num_days : int
        Number of days ot analyze.
    I0 : int, optional
        Initial infected population. The default is 1.
    logger : Logger object, optional
        Logger object. If not given default logging is used.

    Returns
    -------
    result : dict
        Dictionary with fields:
            - name: model name
            - population: Total population
            - data: data Frame with columns
                - S : Susceptible,
                - E : Exposed,
                - I : Infectious (Dying or recovering),
                - Rc : Total Recovered,
                - D : Total deaths
                - R : Removed (R+D),
    """

    second_stage = ph_mix(time_to_die, time_to_recover, fatality_rate)
    total_time = ph_sum(exposed_time, second_stage)
    print(total_time)

    _, _, ae, ne = exposed_time.params()
    _, _, ad, nd = time_to_die.params()
    _, _, ar, nr = time_to_recover.params()

    beta = beta * np.concatenate([np.zeros(ne), np.ones(nd), np.ones(nr)])
    print(beta)

    result = sir_phg(
        name,
        population=population,
        beta=beta,
        infectious_time_distribution=total_time,
        I0=I0,
        num_days=num_days,
        logger=logger,
        report_phases=True,
    )

    exposed_cols, diying_cols, recovering_cols = np.split(result["I-columns"], [ne, ne + nd])
    data = result["data"]
    data["E"] = data[exposed_cols].sum(axis=1)

    _, deaths_cols, recovered_cols = np.split(result["R-columns"], [ne, ne + nd])
    data["D"] = data[deaths_cols].sum(axis=1)
    data["Rc"] = data[recovered_cols].sum(axis=1)
    data["Id"] = data["I"]  # All infected (including E)
    data["I"] = data["I"] - data["E"]  # Actual infectious
    data = data.drop(columns=result["I-columns"] + result["R-columns"])
    result["data"] = data
    result["Name"] = name
    result["population"] = population

    return result



import logging
import os

import pytest
from matplotlib import pyplot as plt
from scipy import stats

from epistoch.seird_ph import seird_ph
from epistoch.sir_g import print_error, report_summary, sir_classical, sir_g
from epistoch.sir_phg import sir_phg
from epistoch.utils.plotting import plot_sir
from pyphase.phase import ph_erlang, ph_expon


def test_sir_g():
    population = 1000
    reproductive_factor = 2.0
    infectious_period_mean = 10
    num_days = 160
    sir_classic = sir_classical(
        name="SIR",
        population=population,
        reproductive_factor=reproductive_factor,
        infectious_period_mean=infectious_period_mean,
        num_days=num_days,
    )
    dist = stats.expon(scale=infectious_period_mean)
    sir_gen = sir_g(
        name="SIR-G",
        population=population,
        reproductive_factor=reproductive_factor,
        infectious_time_distribution=dist,
        num_days=num_days,
    )
    error = print_error(sir_classic, sir_gen)
    report_summary(sir_classic)
    report_summary(sir_gen)
    assert 0.0 == pytest.approx(error, abs=1e-2)


def profile_sir_g():
    import cProfile
    import pstats

    cProfile.run("stochasticSIR(N)", "restats")
    p = pstats.Stats("restats")
    return p


def test_sir_phg_expon(do_plot=False):
    N = 1000
    reproductive_factor = 2.2
    infectious_period_mean = 10
    dist = ph_expon(lambd=1 / infectious_period_mean)
    assert dist.mean() == infectious_period_mean
    gam = 1 / dist.mean()
    beta = reproductive_factor * gam
    sir = sir_classical(
        name="SIR",
        population=N,
        reproductive_factor=reproductive_factor,
        infectious_period_mean=infectious_period_mean,
        num_days=160,
    )
    sir_ph_exp = sir_phg(name="SIR-PH-Expo", population=N, beta=beta, infectious_time_distribution=dist, num_days=160)
    error = print_error(sir, sir_ph_exp)
    if do_plot:
        fig = plot_sir(sir, title="PH-expo-tests")
        plot_sir(sir_ph_exp, N, fig, linestyle="--")
        plt.show()
        assert 0.0 == pytest.approx(error, abs=1e-2)


def test_sir_phg_erlang(do_plot=False):
    N = 1000
    reproductive_factor = 2.2
    infectious_period_mean = 10
    n = 3
    dist = ph_erlang(n=n, lambd=n / infectious_period_mean)
    assert dist.mean() == infectious_period_mean
    gam = 1 / dist.mean()
    beta = reproductive_factor * gam
    num_days = 160
    sir_g_model = sir_g(
        "SIR-G",
        population=N,
        reproductive_factor=reproductive_factor,
        infectious_time_distribution=dist,
        num_days=num_days,
        num_periods=2000,
    )
    sir_ph_erlang = sir_phg("SIR-PHG", population=N, beta=beta, infectious_time_distribution=dist)
    error = print_error(sir_g_model, sir_ph_erlang)
    if do_plot:
        fig = plot_sir(sir_g_model, title="PH-erlang-tests")
        plot_sir(sir_ph_erlang, fig, linestyle="--")
        plt.show()
    assert 0.0 == pytest.approx(error, abs=1e-2)


def test_seird(do_plot=False):
    N = 1000
    beta = 0.2

    exposed_time_mean = 1
    die_time_mean = 7
    recover_time_mean = 12

    exposed_time_exp = ph_expon(mean=exposed_time_mean)
    die_time_exp = ph_expon(mean=die_time_mean)
    recover_time_exp = ph_expon(mean=recover_time_mean)

    exposed_time = ph_erlang(n=10, mean=exposed_time_mean)
    die_time = ph_erlang(n=15, mean=die_time_mean)
    recover_time = ph_erlang(n=8, mean=recover_time_mean)

    fatality_rate = 0.2
    I0 = 10
    num_days = 200

    model_expo = seird_ph(
        "SEIRD-Expo",
        population=N,
        beta=beta,
        exposed_time=exposed_time_exp,
        time_to_die=die_time_exp,
        time_to_recover=recover_time_exp,
        fatality_rate=fatality_rate,
        I0=I0,
        num_days=num_days,
    )
    model_gen = seird_ph(
        "SEIRD-Gen",
        population=N,
        beta=beta,
        exposed_time=exposed_time,
        time_to_die=die_time,
        time_to_recover=recover_time,
        fatality_rate=fatality_rate,
        I0=I0,
        num_days=num_days,
    )
    if do_plot:
        formats = {"S": "b-", "E": "c-", "I": "r-", "R": "g-", "Rc": "y-", "D": "m-"}
        legend_fmt = {"loc": "upper right", "shadow": True, "framealpha": 1.0, "bbox_to_anchor": (1, 1)}
        report_summary(model_expo)
        fig = plot_sir(model_expo, title="SEIRD Expo vs Erlang", formats=formats)

        report_summary(model_gen)
        fig = plot_sir(model_gen, formats=formats, fig=fig, linestyle=":", legend_fmt=legend_fmt)

        plt.show()
        filename = os.path.join("./paper/epistoch/figures/", "SEIRD.pdf")
        print(f"Saving picture in file {os.path.abspath(filename)}")
        fig.savefig(filename, bbox_inches="tight")


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    test_sir_g()

    test_sir_phg_expon()
    test_sir_phg_erlang(True)

    test_seird(do_plot=True)



#!/usr/bin/env python

"""The setup script."""

from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = ["numpy>=1.18", "matplotlib>=3.2.1", "scipy>=1.4", "pandas>=1.0.3", "tqdm>i=4.45.0"]

setup_requirements = [
    "pytest-runner",
]

test_requirements = ["pytest>=3", "pytest-cov"]

setup(
    author="Germán Riaño",
    author_email="griano@germanriano.com",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Epidemics Models with Random Infectious Period",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords="epistoch",
    name="epistoch",
    packages=find_packages(where="src", include=["epistoch", "epistoch.*", "pyphase"]),
    package_dir={"": "src"},
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/griano/epistoch",
    version="0.1.5",
    zip_safe=False,
)



# -*- coding: utf-8 -*-
"""
Created on Sun Apr 12 16:50:30 2020

@author: Germán Riaño
"""

import logging
import math

import numpy as np
import pandas as pd
from scipy import integrate, interpolate, stats
from tqdm import tqdm

import epistoch.utils as utils
from epistoch.sir_g import EPS, _compute_array_error, _sir_deriv, compute_integral, print_error, sir_g
from epistoch.utils.stats import loss_function


def deriv2(
    t, y, beta, gam, I0, times, delta, S_guess, I_guess, survival, pdfs, loss1, dist, method,
):
    # This function does not depend on y, but it depends on t
    # We buid piece-wise linear functions
    if t < min(times) or t > max(times):
        return 0.0, 0.0

    S = interpolate.interp1d(times, S_guess, fill_value="extrapolate")
    I = interpolate.interp1d(times, I_guess, fill_value="extrapolate")
    dSdt = -beta * S(t) * I(t)
    n = int(np.floor(t / delta))
    integral = compute_integral(n, delta, S_guess, I_guess, times, survival, pdfs, loss1, dist, method)
    dIdt = -dSdt - beta * integral - gam * I0 * survival[n]
    # compute_integral(n, delta, S_guess, I_guess, survival, delta_loss)
    # print(f"S'({t})={dSdt}  I'({t})={dIdt}")
    return dSdt, dIdt


def stochasticSIR2(
    population=1000,
    reproductive_factor=2.0,
    disease_time_distribution=stats.expon(scale=10),
    I0=1.0,
    R0=0.0,
    num_days=160,
    num_periods=2000,
    logger=None,
):
    if logger is None:
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.INFO)

    # Total population, N.
    N = population
    dist = disease_time_distribution
    # Normailize to imporve numerical stability
    I0 = I0 / N
    R0 = R0 / N
    # Everyone else, S0, is susceptible to infection initially.
    S0 = 1 - I0 - R0
    # Contact rate, beta, and mean recovery rate, gam, (in 1/days).
    gam = 1 / dist.mean()
    beta = reproductive_factor * gam
    # A grid of time points (in days)
    delta = num_days / num_periods
    times = np.linspace(start=0.0, stop=num_days, num=num_periods + 1)
    survival = dist.sf(times)
    pdfs = dist.pdf(times)
    if pdfs[0] == math.inf:
        pdfs[0] = dist.pdf(dist.ppf(0.01))
    survival = dist.sf(times)
    if survival[0] < 1.0 - EPS:
        logging.warning(f"Survival function does not start in 1:  {survival}")
    loss_fun = loss_function(dist)
    loss1 = loss_fun(times)

    S = np.empty(num_periods + 1)
    S[:] = np.nan
    I = np.empty(num_periods + 1)
    I[:] = np.nan
    S[0] = S0
    I[0] = I0

    # Initial conditions vector
    y0 = [S0, I0]

    max_iterations = 100
    # Integrate the SIR equations over the time grid, t.
    ret = integrate.odeint(_sir_deriv, y0, times, args=(beta, gam), tfirst=True)
    old_y = ret.T
    max_diff = 1e-3
    alpha = 0.5
    for iteration in tqdm(range(max_iterations)):
        S_guess, I_guess = old_y
        # Integrate again, with the integral, fixing guess
        args = (
            beta,
            gam,
            I0,
            times,
            delta,
            S_guess,
            I_guess,
            survival,
            pdfs,
            loss1,
            dist,
            "loss",
        )
        ret2 = integrate.odeint(func=deriv2, y0=y0, t=times, args=args, tfirst=True)
        y = ret2.T
        S, I = y
        diff = _compute_array_error("I", I_guess, I, do_print=False)
        logger.debug(f"Iteration {iteration:d}: {diff:.4%}")
        if diff < max_diff:
            break
        old_y = alpha * y + (1 - alpha) * old_y

    # Report one point per day, and de-normalize
    days = np.linspace(0, num_days, num_days + 1)
    S = N * interpolate.interp1d(times, S)(days)
    I = N * interpolate.interp1d(times, I)(days)
    R = N - S - I
    return pd.DataFrame(data={"Day": days, "S": S, "I": I, "R": R}).set_index("Day")


def test_sir_g2(num_periods=2000):
    N = 1000
    dist = utils.stats.get_gamma(10, 2)
    model1 = sir_g(population=N, infectious_time_distribution=dist, num_periods=num_periods)
    model2 = stochasticSIR2(population=N, disease_time_distribution=dist, num_periods=num_periods)
    model1.plot()
    model2.plot()
    print_error(model1, model2, N)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    test_sir_g2()



# -*- coding: utf-8 -*-
"""
Created on Wed Apr 15 15:34:26 2020

@author: Germán Riaño
"""

import logging
import os
from timeit import timeit

import matplotlib.pyplot as plt
import numpy as np
import scipy

import epistoch.utils as utils
from epistoch.sir_g import print_error, report_summary, sir_classical, sir_g
from epistoch.utils.plotting import plot_IR, plot_sir


def compare_models(name, dist, N=1000000, I0=1, num_days=100, R0=2.25, do_plots=True):
    name1 = "SIR"
    SIR_classic = sir_classical(
        name=name1, population=N, I0=I0, reproductive_factor=R0, infectious_period_mean=dist.mean(), num_days=num_days,
    )

    name2 = "SIR-G"
    SIR_general = sir_g(
        name=name2,
        population=N,
        I0=I0,
        num_days=num_days,
        num_periods=2000,
        reproductive_factor=R0,
        infectious_time_distribution=dist,
        method="loss",
    )
    report_summary(SIR_classic)
    report_summary(SIR_general)
    print_error(SIR_classic, SIR_general)

    if do_plots:
        fig = plot_sir(SIR_classic, title=name + ": SIR and SIR-G models")
        plot_sir(SIR_general, fig, linestyle="--")
        plt.show()
        filename = os.path.join("./paper/epistoch/figures/", name + "-SIR-comp.pdf")
        print(f"Saving picture in file {os.path.abspath(filename)}")
        fig.savefig(filename, bbox_inches="tight")

        fig2 = plot_IR(SIR_classic, title=name + ": I, R as function of S")
        plot_IR(SIR_general, fig2, linestyle="--")
        plt.show()
        filename = os.path.join("./paper/epistoch/figures/", name + "-IR-comp.pdf")
        print(f"Saving picture in file {os.path.abspath(filename)}")
        fig2.savefig(filename, bbox_inches="tight")
        print("Done")


def variance_analysis(
    N=1000000, I0=1, num_days=100, R0=2.25, infectious_period_mean=4.4, cvs=[0.5, 1.0, 2.0],
):

    dists = {}
    models = {}
    print("Computing Classical")
    sir_classic = sir_classical(
        name="SIR",
        population=N,
        I0=I0,
        reproductive_factor=R0,
        infectious_period_mean=infectious_period_mean,
        num_days=num_days,
    )
    fig = plot_sir(sir_classic, formats={"I": "r-"}, title="SIR-G with Different Distributions", linewidth=3,)

    print("Computing Constant SIR-G")
    dist = utils.stats.constant(infectious_period_mean)
    sir_constant = sir_g(
        name="Const",
        population=N,
        I0=I0,
        num_days=num_days,
        num_periods=2000,
        reproductive_factor=R0,
        infectious_time_distribution=dist,
        method="loss",
    )
    fig = plot_sir(sir_constant, fig, formats={"I": "b-."}, linewidth=3)

    dists["gamma"] = utils.stats.get_gamma
    dists["lognorm"] = utils.stats.get_lognorm

    for cv in cvs:
        for dist_name, dist_getter in dists.items():
            infectious_period_sd = cv * infectious_period_mean
            dist = dist_getter(infectious_period_mean, infectious_period_sd)
            mod_name = f"{dist_name}-{cv:.2}"
            print(f"Running {mod_name}")
            num_days = round(100 * max(1, cv))
            models[(dist, cv)] = sir_g(
                name=mod_name,
                population=N,
                I0=I0,
                num_days=num_days,
                num_periods=2000,
                reproductive_factor=R0,
                infectious_time_distribution=dist,
                method="loss",
            )
            #            models[(dist, cv)].name = mod_name
            plot_sir(models[(dist, cv)], fig, formats={"I": ""}, linestyle="--")
    plt.show()

    filename = os.path.join("./paper/epistoch/figures/", "Variance-Analysis.pdf")
    print(f"Saving picture in file {os.path.abspath(filename)}")
    fig.savefig(filename, bbox_inches="tight")

    return models


if __name__ == "__main__":
    # logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)
    # logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)
    logging.basicConfig()
    compare_models("DIVOC", scipy.stats.norm(loc=4.5, scale=1))
    variance_analysis()



# -*- coding: utf-8 -*-
"""
Created on Sat Apr 25 11:55:02 2020

@author: Germán Riaño
"""
import numpy as np
from numpy import matlib as ml
from scipy import linalg
from scipy.special import binom
from scipy.stats import rv_continuous


class PhaseTypeGen(rv_continuous):
    def _set_pars(self, alpha, A):
        alpha = ml.mat(alpha)
        A = ml.mat(A)
        self.alpha = alpha
        self.A = A
        self.alphaAi = None  # alpha * A^(-1), computed lazily
        self.erd = None  # Equilibrium residual distribution

    def _cdf1(self, x):
        return 1.0 - np.sum(self.alpha @ linalg.expm(self.A.A * x), axis=1) if x > 0 else 0.0

    def _cdf(self, x):
        return np.vectorize(self._cdf1)(x)

    def _pdf1(self, x):
        res = self.alpha @ linalg.expm(self.A.A * x) @ self._get_a() if x >= 0 else 0.0
        return res

    def _pdf(self, x):
        return np.vectorize(self._pdf1)(x)

    def _moments(self, k=None):
        if k is None:
            k = 2
        moms = list()
        left = self.alpha
        for i in range(1, k + 1):
            left = -i * self._solve_vector_A(left, self.A)
            moms.append(np.sum(left.flatten()))
        return moms

    def _stats(self):
        # Mean(‘m’), variance(‘v’), skew(‘s’), and/or kurtosis(‘k’).
        moments = self._moments(4)
        moments = np.concatenate(([1.0], moments))
        mean = moments[1]
        centered_moments = {
            n: sum([binom(n, k) * moments[k] * (-mean) ** (n - k) for k in range(0, n + 1)]) for n in range(1, 5)
        }
        var = centered_moments[2]
        sd = np.sqrt(var)
        skew = centered_moments[3] / sd ** 3
        # kurtosis is fourth central moment / variance**2 - 3
        kurt = centered_moments[4] / (var * var) - 3.0
        return mean, var, skew, kurt

    def params(self):
        return self.alpha, self.A, self._get_a(), self.A.shape[0]

    def _get_a(self):
        return -np.sum(self.A, axis=1)

    def _solve_vector_A(self, vector, A):
        # Computes vector * A^(1), without actually inverting the matrix, by solving x * A = vector
        return np.linalg.solve(A.T, vector.T).T

    def _get_alpha_Ai(self):
        # Computes vector * A^(1), without actually inverting the matrix, by solving x * A = vector
        if self.alphaAi is None:
            self.alphaAi = self._solve_vector_A(self.alpha, self.A)
        return self.alphaAi

    def _get_erd(self):
        if self.erd is None:
            gamma = 1 / self.mean()
            self.erd = phase(-gamma * self._get_alpha_Ai(), self.A)
        return self.erd

    def loss1(self, x):
        # First order loss function
        return self.mean() * self._get_erd().sf(x)

    def equilibrium_pi(self):
        # Solution to pi (A + alpha * a)=0, pi * one = 1
        n = self.A.shape[0]
        a = self._get_a()
        matrix = self.A + a @ self.alpha + np.ones((n, n))
        return np.linalg.solve(matrix.T, np.ones(n).T).T

    def _pre_append(self, stg, pre):
        lines = stg.splitlines()
        result = [lines[0]]
        for line in lines[1:]:
            result.append(pre + line)
        return "\n".join(result)

    def __str__(self):
        options = {"precision": 2}
        alpha_stg = np.array2string(self.alpha, **options)
        A_stg = np.array2string(self.A, **options)
        return "PhaseType: \n" f"  alpha = {alpha_stg}\n" f"  A     = {self._pre_append(A_stg,'          ')}"

    def __repr__(self):
        return f"PH({self.alpha.__repr__()}, {self.A.__repr__()})"


def phase(alpha, A):
    generator = PhaseTypeGen(name="phase")
    frozen = generator.__call__()
    frozen.dist._set_pars(alpha, A)
    frozen.equilibrium_pi = frozen.dist.equilibrium_pi
    frozen.loss1 = frozen.dist.loss1
    frozen.params = frozen.dist.params
    return frozen


def ph_expon(lambd=None, mean=None):
    return ph_erlang(1, lambd, mean)


def ph_erlang(n, lambd=None, mean=None):
    """
    Builds an Erlang(n,lmbda).
    If mean is provided then lambda = n/mean

    Parameters
    ----------
    n : int
        The order of theis Erlang
    lambd : float, optional
        Provide only one between lambd or mean
    mean : float, optional
        Provide only one between lambd or mean

    Returns
    -------
    TYPE
        PH representation for an Erlang

    """
    if lambd is None and mean is None:
        raise ValueError('"You must provide one between mean and lambda')
    lambd = lambd if lambd is not None else n / mean
    A = -lambd * np.eye(n)
    for i in range(n - 1):
        A[i, i + 1] = lambd
    alpha = np.zeros(n)
    alpha[0] = 1
    return phase(alpha, A)


def _z(n1, n2):
    # quiclky generates zeros
    return np.zeros((n1, n2))


def ph_sum(ph1, ph2):
    alpha1, A1, a1, n1 = ph1.params()
    alpha2, A2, a2, n2 = ph2.params()
    alpha = np.block([alpha1, _z(1, n2)])
    A = np.block([[A1, a1 * alpha2], [_z(n2, n1), A2]])
    return phase(alpha, A)


def ph_mix(ph1, ph2, p1):
    alpha1, A1, a1, n1 = ph1.params()
    alpha2, A2, a2, n2 = ph2.params()
    alpha = np.block([p1 * alpha1, (1 - p1) * alpha2])
    A = np.block([[A1, _z(n1, n2)], [_z(n2, n1), A2]])
    return phase(alpha, A)



# -*- coding: utf-8 -*-
"""
Created on Sun Apr 26 11:59:01 2020

@author: Germán Riaño, PhD
"""

import numpy as np
from scipy.stats import rv_continuous


class mixrv_gen(rv_continuous):
    def __init__(self, *args, **kwargs):
        super(mixrv_gen, self).__init__(*args, **kwargs)

    def _set_pars(self, ps, vs):
        n = max(ps.shape)
        ps.shape = (n, 1)
        self.ps = ps
        self.vs = vs

    def _cdf(self, x):
        ps = self.ps
        vs = self.vs
        return np.sum(ps * np.array([v.cdf(x) for v in vs]), axis=0)

    def _pdf(self, x):
        ps = self.ps
        vs = self.vs
        return np.sum(ps * np.array([v.pdf(x) for v in vs]), axis=0)

    def _munp(self, n):
        ps = np.transpose(self.ps)
        vs = self.vs
        return np.sum(ps * np.array([v.moment(n) for v in vs]))


def mix_rv(ps, vs):
    generator = mixrv_gen("mixture")
    frozen = generator.__call__()
    frozen.dist._set_pars(ps, vs)
    return frozen


def assert_dist_identical(dist1, dist2, name=None):
    print("Comparing Distributions " + (name if name is not None else ""))
    print("  Testing stats")
    np.testing.assert_allclose(dist1.stats(moments="mvsk"), dist2.stats(moments="mvsk"))
    print("  Testing CDF")
    p1 = np.linspace(0.0, 0.99)
    x1 = dist1.ppf(p1)
    p1 = np.concatenate(([0, 0, 0], p1))
    x1 = np.concatenate(([-2, -1, 0], x1))
    p2 = dist2.cdf(x1)
    np.testing.assert_allclose(p1, p2)
    print("  Testing PDF")
    np.testing.assert_allclose(dist1.pdf(x1), dist2.pdf(x1))
